
#include <phg.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>

typedef struct {
    /* the levelset function, defined either by DOFs or function pointers */
    DOF		*ls, *ls_grad;
    FLOAT	(*frac)[2];	/* volume fractions of -/+ parts in an elem */
    BOOLEAN	dof_owned;	/* TRUE => owns ls, ls_grad */
} XFEM_INFO;

XFEM_INFO *
phgXFEMInit(GRID *g, FUNC_3D ls_func, int ls_order, FUNC_3D ls_grad)
{
    XFEM_INFO *xi = phgCalloc(1, sizeof(*xi));

    xi->ls = phgDofNew(g, DOF_ANALYTIC, 1, "ls", ls_func);
    xi->ls_grad = phgDofNew(g, DOF_ANALYTIC, 1, "ls_grad", ls_grad);;
    if (ls_order >= 0) {
	phgDofSetPolyOrder(xi->ls, ls_order);
	phgDofSetPolyOrder(xi->ls_grad, ls_order == 0 ? 0 : ls_order - 1);
    }
    xi->frac = phgCalloc(g->nelem, sizeof(*xi->frac));
    xi->dof_owned = TRUE;

    phgQuadInterfaceMarkGrid(xi->ls, xi->ls_grad);

    return xi;
}

void
phgXFEMFree(XFEM_INFO **xi)
{
    if (xi == NULL || *xi == NULL)
	return;
    phgFree((*xi)->frac);
    if ((*xi)->dof_owned) {
	phgDofFree(&(*xi)->ls);
	phgDofFree(&(*xi)->ls_grad);
    }
    phgFree(*xi);
    *xi = NULL;
}

FLOAT
phgXFEMDot(XFEM_INFO *xi, DOF *um, DOF *up, DOF *vm, DOF *vp)
/* computes L2 dot-product (u,v), where
 *	u:=um and v:=vm in \Omega^-,
 * and
 *	u:=up and v:=vp in \Omega^+.
 */
{
    int dim = DofDim(um), Q_um, Q_vm, Q_up, Q_vp;
    FLOAT dot = 0.0;
    GRID *g = xi->ls->g;
    ELEMENT *e;
    QCACHE *qc = phgQCNew(QD_PHG, um);

    Q_um = phgQCAddFEFunction(qc, um);
    Q_vm = um == vm ? Q_um : phgQCAddFEFunction(qc, vm);

    Q_up = phgQCAddFEFunction(qc, up);
    Q_vp = up == vp ? Q_up : phgQCAddFEFunction(qc, vp);

    assert(dim == DofDim(up) && dim == DofDim(vm) && dim == DofDim(vp));

#if USE_OMP
#pragma omp parallel for private(e) reduction(+:dot)
#endif	/* USE_OMP */
    ForAllElementsBegin(g, e) {
	int i, j;
	FLOAT *rule_m = NULL, *rule_p = NULL;
	int order;	/* quadrature order */
	i = DofTypeOrder(um, e);
	if (i < DofTypeOrder(vm, e))
	    i = DofTypeOrder(vm, e);
	j = DofTypeOrder(up, e);
	if (j < DofTypeOrder(vp, e))
	    j = DofTypeOrder(vp, e);
	order = i + j + (e->mark == 0 ? 3 : 0);
	phgQuadInterface(e->mark == 0 ? xi->ls : NULL, xi->ls_grad, e, order,
			e->mark <= 0 ? &rule_m : NULL, NULL,
			e->mark >= 0 ? &rule_p : NULL);
	phgQCSetRule(qc, rule_m, -1.);
	dot += phgQCIntegrate(qc, e->index, Q_um, 0, qc, e->index, Q_vm, 0);
	phgFree(rule_m);

	phgQCSetRule(qc, rule_p, -1.);
	dot += phgQCIntegrate(qc, e->index, Q_up, 0, qc, e->index, Q_vp, 0);
	phgFree(rule_p);
    } ForAllElementsEnd

    phgQCFree(&qc);


#if USE_MPI
    if (g->nprocs > 1) {
	FLOAT d = dot;
	MPI_Allreduce(&d, &dot, 1, PHG_MPI_FLOAT, PHG_SUM, g->comm);
    }
#endif	/* USE_MPI */

    return dot;
}

static void
process_empty_rows(SOLVER *solver)
/* process empty row */
{
    int i;
    MAT *A = phgSolverGetMat(solver);
    MAT_ROW *row;

    phgMatAssemble(A);
    phgMatDisassemble(A);
    assert(A->type == PHG_UNPACKED);
    for (i = 0; i < A->rmap->nlocal; i++) {
	row = A->rows + i;
	if (row->ncols == 0) {
	    /* empty row */
	    row->ncols = row->alloc = 1;
	    phgFree(row->cols);
	    row->cols = phgAlloc(sizeof(*row->cols));
	    row->cols[0] = i;
	    phgFree(row->data);
	    row->data = phgAlloc(sizeof(*row->data));
	    row->data[0] = 1.0;
	}
    }
    phgMatAssemble(A);
    phgMatDestroy(&A);
}

/*---------------------------------------------------------------------------
 * This program solves the following interface problem using XFEM:
 * 	-\div (a \grad u) = f,	in \Omega^- \cup \Omega^+,
 * 	u = g,			on \partial\Omega,
 * 	[u] = g_D,		on \Gamma,
 * 	[a\grad u] = g_N,	on \Gamma,
 * Where
 * 	\Omega^- := \Omega \cap \{x: L(x)<0\},
 * 	\Omega^+ := \Omega \cap \{x: L(x)>0\},
 * 	\Gamma	 := \Omega \cap \{x: L(x)=0\},
 * 	a := COEEF1 for x\in\Omega^- and COEEF2 for x\in\Omega^+
 * L(x) is the level-set function defined by ls_func()
 *
 * The analytic test solution is given by func_u1() in \Omega^- and func_u2()
 * in \Omega^+.
 *
 * $Id: interface.c,v 1.161 2021/06/25 13:00:59 zlb Exp $
 *---------------------------------------------------------------------------*/

/* The coefficients */
static FLOAT coeff1 = 0.5, coeff2 = 2.0;

/* The IP parameters */
static FLOAT beta = -1.0, gamma0 = 1.0, gamma1 = 1.0;

static BOOLEAN dump_mat = FALSE, dump_rhs = FALSE;

/* analytic solution */
static void
func_u1(FLOAT x, FLOAT y, FLOAT z, FLOAT *value)
{
    *value = Exp(x * y * z);
}

static void
func_u2(FLOAT x, FLOAT y, FLOAT z, FLOAT *value)
{
    *value = Sin(x + y + z);
}

static void
func_grad_u1(FLOAT x, FLOAT y, FLOAT z, FLOAT *values)
{
    *values++ = y * z * Exp(x * y * z);
    *values++ = x * z * Exp(x * y * z);
    *values++ = x * y * Exp(x * y * z);
}

static void
func_grad_u2(FLOAT x, FLOAT y, FLOAT z, FLOAT *values)
{
    *values++ = Cos(x + y + z);
    *values++ = Cos(x + y + z);
    *values++ = Cos(x + y + z);
}

/* right hand side */
static void
func_f1(FLOAT x, FLOAT y, FLOAT z, FLOAT *value)
{
    *value = -(x * x * y * y + x * x * z * z + y * y * z * z)
	      * Exp(x * y * z) * coeff1;
}

static void
func_f2(FLOAT x, FLOAT y, FLOAT z, FLOAT *value)
{
#if 0
#warning Test with continuous solution and discontinuous coefficient
#define func_u2 func_u1
#define func_grad_u2 func_grad_u1
    func_f1(x, y, z, value);
    *value *= coeff2 / coeff1;
#else
    *value = 3. * Sin(x + y + z) * coeff2;
#endif
}

/* the levelset function */
static FLOAT xc = 0.5, yc = 0.5, zc = 0.5, r = 0.25;

static void
ls_func(FLOAT x, FLOAT y, FLOAT z, FLOAT *value)
/* level set function */
{
    x -= xc; y -= yc; z -= zc;
    *value = x * x + y * y + z * z - r * r;
}

static void
ls_grad(FLOAT x, FLOAT y, FLOAT z, FLOAT *grad)
/* level set function */
{
    x -= xc; y -= yc; z -= zc;
    grad[0] = x + x;
    grad[1] = y + y;
    grad[2] = z + z;
}

#if 0
/* FIXME: this is noticebly faster than calling libphg, e.g.,
 * 	./interface -dof_type DG1 -refine 12 -solver mumps
 * (IPO?) */
# include "../src/quad-cache.c"
#endif

/* CACHE_MODE == 2: cache bases, funtions, A*functions, A*grad and projections
 * CACHE_MODE == 1: cache bases, functions, A*functions and A*grad
 * CACHE_MODE == 0: cache bases, functions and projections */
#ifndef CACHE_MODE
# define CACHE_MODE 0
#endif

/* function numbers in the QCACHE objects */
int Q_f, Q_gD, Q_gN;			/* All MODEs */
#if CACHE_MODE
int Q_AgN, Q_AGrad;			/* MODE == 1 or 2 */
# if CACHE_MODE == 2
int Q_AgN_n, Q_GRAD_n, Q_AGrad_n;	/* MODE == 2 */
# endif	/* CACHE_MODE == 2 */
#else
int Q_gN_n, Q_GRAD_n;			/* MODE == 0 */
#endif	/* CACHE_MODE */

static void
do_face(XFEM_INFO *xi, SOLVER *solver, FLOAT *G,
	QCACHE *qc,  ELEMENT *e,  int face,  int dof_id,  FLOAT coeff1,
	QCACHE *qc1, ELEMENT *e1, int face1, int dof_id1, FLOAT coeff2)
/* This function computes integrations on a surface area, which can be:
 *
 *   1) an element face, in this case:
 * 	e != e1 && dof_id == dof_id1, e and e1 are both non interface elements
 *
 *   2) part of an element face, in this case:
 *   	e != e1 && dof_id == dof_id1, e and e1 are both interface elements
 *
 *   3) \Gamma \cap e, in this case:
 *	e == e1, dof_id = 0, dof_id1 = 1, and e is an interface element
 *
 * For cases 1) and 2), the normal vector stored in qc is the outward normal
 * vector of e, and e1 can be NULL (boundary face).
 *
 * For the case 3) the normal vectors stored in qc points from \Omega^- to
 * \Omega^+.
 *
 * The argument G, if not NULL, is used to bring back the actual penalty
 * parameters (G[0]=G0 and G[1]=G1) for use by the calling function,
 * to ensure the same penalty parameters are used.
 */
{
    DOF *u_h, *u1_h;
    int n, n1, p;   /* p = polynomial order */
    INT I, J;
    int i, j;
    FLOAT val;
    FLOAT G0, G1, h, a;     /* G0 = gamma0*p^2/h, G1 = gamma1*h/p^2 */
    enum {INTERIOR, DIRICHLET, NEUMANN} bdry_flag;

    if (phgQCGetNP(qc) == 0)
	return;

    assert((dof_id != dof_id1) == (e == e1));

    u_h = qc->fe;
    u1_h = qc1->fe;

    n = qc->qd->n_bas(qc, e->index);
    p = DofTypeOrder(u_h, e);

    h = e == e1 ? phgGeomGetDiameter(xi->ls->g, e) :
		  phgGeomGetFaceDiameter(xi->ls->g, e, face);

    if (e1 == NULL) {
	/* boundary face */
	bdry_flag = DIRICHLET;	/* TODO: Neumann boundary */
	n1 = 0;
	G0 = gamma0 * p * p / h;
	G1 = gamma1 * h / (p * (FLOAT)p);
	/* RHS */
	for (i = 0; i < n; i++) {
	    I = phgSolverMapE2G(solver, dof_id, e, i);

	    /* -\beta\int_\Gamma_D g (A\grad v).n */
#if CACHE_MODE == 2
	    a = phgQCIntegrateFace(qc, e->index, face, Q_gD,      PROJ_NONE, 0,
				   qc, e->index, face, Q_AGrad_n, PROJ_NONE, i);
#elif CACHE_MODE == 1
	    a = phgQCIntegrateFace(qc, e->index, face, Q_gD,    PROJ_NONE, 0,
				   qc, e->index, face, Q_AGrad, PROJ_DOT, i);
#else
	    a = phgQCIntegrateFace(qc, e->index, face, Q_gD,     PROJ_NONE, 0,
				   qc, e->index, face, Q_GRAD_n, PROJ_NONE, i)
		* coeff1;
#endif
	    val = -beta * a;

	    /* G0 \int_\Gamma_D g v */
	    a = phgQCIntegrateFace(qc, e->index, face, Q_gD, PROJ_NONE, 0,
				   qc, e->index, face, Q_BAS, PROJ_NONE, i);
	    val += G0 * a;

	    phgSolverAddGlobalRHSEntry(solver, I, val);
	}
    }
    else {
	bdry_flag = INTERIOR;
	n1 = qc->qd->n_bas(qc1, e1->index);
	i = DofTypeOrder(u1_h, e1);
	if (p < i)
	    p = i;
	G0 = gamma0 * p * (FLOAT)p / h;
	G1 = gamma1 * h / (p * (FLOAT)p);
    }

    if (G != NULL) {
	/* set penalizing parameters for use by the calling function */
	G[0] = G0;
	G[1] = G1;
    }

    /* The following macros depend on the assertion */

#define Sel(i, o, o1)	(i < n ? o : o1)
#define Qc(i)	Sel(i, qc, qc1)
#define Ele(i)	Sel(i, e, e1)
#define Fac(i)	Sel(i, face, face1)
#define Bas(i)	Sel(i, i, i - n)
#define Dof(i)	Sel(i, dof_id, dof_id1)
#define Coe(i)	Sel(i, coeff1, coeff2)
#define Quad(fid1, proj1, i1, fid2, proj2, i2) phgQCIntegrateFace( \
		Qc(i1), Ele(i1)->index, Fac(i1), fid1, proj1, Bas(i1), \
		Qc(i2), Ele(i2)->index, Fac(i2), fid2, proj2, Bas(i2))

    /* loop on {basis funcs in e} \cup {basis funcs in e1} */
    for (i = 0; i < n + n1; i++) {
	I = phgSolverMapE2G(solver, Dof(i), Ele(i), Bas(i));
	/* loop on {basis funcs in e} \cup {basis funcs in e1} */
	for (j = 0; j < n + n1; j++) {
	    J = phgSolverMapE2G(solver, Dof(j), Ele(j), Bas(j));

	    /* skip jumps for interior face and continuous element */
	    if (DofFESpace(u_h) == FE_H1 && e != e1 && e1 != NULL) {
		assert(bdry_flag == INTERIOR && dof_id == dof_id1);
		continue;
	    }

	    /*-----------------------------------------------------------------
	     * Note if the normal vector is reversed, the sign of [.] changes,
	     * while the sign of {.} is unaffected.
	     *----------------------------------------------------------------*/

	    /* -\int {A\grad u}.n [v] (i<=>v, j<=>u, n<=>e) */
#if CACHE_MODE == 2
	    a = Quad(Q_AGrad_n, PROJ_NONE, j, Q_BAS, PROJ_NONE, i);
#elif CACHE_MODE == 1
	    a = Quad(Q_AGrad,   PROJ_DOT,  j, Q_BAS, PROJ_NONE, i);
#else
	    a = Quad(Q_GRAD_n,  PROJ_NONE, j, Q_BAS, PROJ_NONE, i) * Coe(j);
#endif	/* CACHE_PROJECTION */
	    if (bdry_flag == INTERIOR)
		a *= (i < n ? 0.5 : -0.5);
	    val = -a;

	    /* -\int \beta [u]{A\grad v}.n (note: i<=>v, j<=>u, n<=>e) */
#if CACHE_MODE == 2
	    a = Quad(Q_BAS, PROJ_NONE, j, Q_AGrad_n, PROJ_NONE, i);
#elif CACHE_MODE == 1
	    a = Quad(Q_BAS, PROJ_NONE, j, Q_AGrad,   PROJ_DOT, i);
#else
	    a = Quad(Q_BAS, PROJ_NONE, j, Q_GRAD_n,  PROJ_NONE, i) * Coe(i);
#endif	/* CACHE_PROJECTION */
	    if (bdry_flag == INTERIOR)
		a *= (j < n ? 0.5 : -0.5);
	    val += -beta * a;

#if 0
#warning test code!
# define TEST_QUAD 1
QCACHE *q0 = qc;
if (e1 != e && e1 != NULL) {
    FLOAT *rule;
    assert(qc == qc1);
    qc = qc1 = phgQCClone(q0);
    phgQuadInterfaceFace(NULL, xi->ls_grad, e, face, p + p, &rule, NULL, NULL);
    phgQCSetRule(qc, rule, -1.);
}
#else
#define TEST_QUAD 0
Unused(xi);
#endif
	    /* \int G0 [u][v] (i<=>v, j<=>u, n<=>e) */
	    a = Quad(Q_BAS, PROJ_NONE, j, Q_BAS, PROJ_NONE, i);
	    if (bdry_flag == INTERIOR && (i < n) != (j < n))
		a = -a;
	    val += G0 * a;

	    if (bdry_flag != DIRICHLET) {
		/* \int G1 [A\grad u].n [A\grad v].n */
#if CACHE_MODE == 2
		a = Quad(Q_AGrad_n, PROJ_NONE, j, Q_AGrad_n, PROJ_NONE, i);
#elif CACHE_MODE == 1
		a = Quad(Q_AGrad,   PROJ_DOT,  j, Q_AGrad,   PROJ_DOT, i);
#else
		a = Quad(Q_GRAD_n,  PROJ_NONE, j, Q_GRAD_n,  PROJ_NONE, i)
			* Coe(j) * Coe(i);
#endif	/* CACHE_PROJECTION */
		if (bdry_flag == INTERIOR && (i < n) != (j < n))
		    a = -a;
		val += G1 * a;
	    }
#if TEST_QUAD
if (qc != q0) {
    phgQCFree(&qc);
    qc = qc1 = q0;
}
#endif

	    phgSolverAddGlobalMatrixEntry(solver, I, J, val);
	}
    }
#undef Sel
#undef Qc
#undef Ele
#undef Fac
#undef Bas
#undef Dof
#undef Coe
#undef Quad
}

static void
build_linear_system(XFEM_INFO *xi, SOLVER *solver, DOF *u1_h, DOF *u2_h,
		    DOF *f1_h, DOF *f2_h)
{
    GRID *g = xi->ls->g;
    QCACHE *qm, *qp;
    ELEMENT *e;
    /* statistics on the minimum fractions of the subdomains */
    double minfrac[2], tmp[2][phgMaxThreads];
    INT ecnts[3], etmp[3][phgMaxThreads];	/* ecnts: 0=- 1=0 2=+ */
    int ii;
    
    assert(u1_h->dim == 1 && u2_h->dim == 1);

    phgPrintf("    CACHE_MODE = %d\n", CACHE_MODE);

    qm = phgQCNew(QD_PHG, u1_h);
    Q_f = phgQCAddFEFunction(qm, f1_h);				/* f */
    Q_gD = phgQCAddXYZFunction(qm, func_u1, 1);			/* gD */
    Q_gN = phgQCAddXYZFunction(qm, func_grad_u1, 3);		/* gN */

    qp = phgQCNew(QD_PHG, u2_h);
    /* The following must be in exactly the same order as above */
    phgQCAddFEFunction(qp, f2_h);				/* f  */
    phgQCAddXYZFunction(qp, func_u2, 1);			/* gD */
    phgQCAddXYZFunction(qp, func_grad_u2, 3);			/* gN */

#if CACHE_MODE
    Q_AGrad = phgQCAddConstantCoefficient(qm, &coeff1, 1, Q_GRAD);/* AGrad */
    Q_AgN = phgQCAddConstantCoefficient(qm, &coeff1, 1, Q_gN);	/* AgN */

    phgQCAddConstantCoefficient(qp, &coeff2, 1, Q_GRAD);	/* AGrad */
    phgQCAddConstantCoefficient(qp, &coeff2, 1, Q_gN);		/* AgN */

# if CACHE_MODE == 2
    /* normal projections */
    Q_AgN_n = phgQCAddProjection(qm, PROJ_DOT, Q_AgN);		/* AgN.n */
    Q_GRAD_n = phgQCAddProjection(qm, PROJ_DOT, Q_GRAD);	/* Grad.n */
    Q_AGrad_n = phgQCAddProjection(qm, PROJ_DOT, Q_AGrad);	/* AGrad.n */

    phgQCAddProjection(qp, PROJ_DOT, Q_AgN);			/* AgN.n */
    phgQCAddProjection(qp, PROJ_DOT, Q_GRAD);			/* Grad.n */
    phgQCAddProjection(qp, PROJ_DOT, Q_AGrad);			/* AGrad.n */
# endif	/* CACHE_MODE == 2 */
#else
    Q_gN_n = phgQCAddProjection(qm, PROJ_DOT, Q_gN);		/* gN.n */
    Q_GRAD_n = phgQCAddProjection(qm, PROJ_DOT, Q_GRAD);	/* Grad.n */

    phgQCAddProjection(qp, PROJ_DOT, Q_gN);			/* gN.n */
    phgQCAddProjection(qp, PROJ_DOT, Q_GRAD);			/* Grad.n */
#endif	/* CACHE_MODE */

    for (ii = 0; ii < phgMaxThreads; ii++) {
	tmp[0][ii] = tmp[1][ii] = 2.0;
	etmp[0][ii] = etmp[1][ii] = etmp[2][ii] = 0;
    }

#if USE_OMP
#pragma omp parallel for private(e)
#endif	/* USE_OMP */
    ForAllElementsBegin(g, e) {
	FLOAT *rule_m = NULL, *rule_0 = NULL, *rule_p = NULL;
	int N = qm->qd->n_bas(qm, e->index);
	INT I, J, eno = e->index;
	int i, j, k, face, order;
	FLOAT val;

	/* construct quadrature rules for the subdomains and the interface */
	i = DofTypeOrder(u1_h, e);
	j = DofTypeOrder(u2_h, e);
	order = 2 * (i > j ? i : j);
	if (e->mark == 0) {
	    phgQuadInterface(xi->ls, xi->ls_grad, e, order + 3,
				&rule_m, &rule_0, &rule_p);
	    if (phgQuadInterfaceRuleInfo(rule_p, NULL, NULL, NULL) == 0)
		e->mark = -1;
	    else if (phgQuadInterfaceRuleInfo(rule_m, NULL, NULL, NULL) == 0)
		e->mark = 1;
	}
	if (e->mark < 0) {
	    phgFree(rule_m);
	    phgQuadInterface(NULL, xi->ls_grad, e, order /*- 2*/,
			     &rule_m, NULL, NULL);
	    xi->frac[eno][0] = 1.0;
	    xi->frac[eno][1] = 0.0;
	    if (tmp[0][phgThreadId] > 1.0)
		tmp[0][phgThreadId] = 1.0;
	    etmp[0][phgThreadId]++;
	}
	else if (e->mark > 0) {
	    phgFree(rule_p);
	    phgQuadInterface(NULL, xi->ls_grad, e, order/* - 2*/,
			     NULL, NULL, &rule_p);
	    xi->frac[eno][0] = 0.0;
	    xi->frac[eno][1] = 1.0;
	    if (tmp[1][phgThreadId] > 1.0)
		tmp[1][phgThreadId] = 1.0;
	    etmp[2][phgThreadId]++;
	}
	else {
	    FLOAT *pw;
	    for (k = 0; k < 2; k++) {
		j = phgQuadInterfaceRuleInfo( k == 0 ? rule_m : rule_p,
							NULL, &pw, NULL);
		for (i = 0, val = 0.; i < j; i++, pw += Dim + 1)
		    val += pw[Dim];
		xi->frac[eno][k] = (val /= phgGeomGetVolume(g, e));
		if (val > 0. && tmp[k][phgThreadId] > val)
		    tmp[k][phgThreadId] = val;
	    }
	    etmp[1][phgThreadId]++;
	}

	/*==================== volume integrals ====================*/

	phgQCSetRule(qm, rule_m, -1.);
	phgQCSetRule(qp, rule_p, -1.);

	for (k = 0; k < 2; k++) {
	    QCACHE *q;
	    if ((k == 0 && e->mark > 0) || (k == 1 && e->mark < 0))
		continue;
	    q = k == 0 ? qm : qp;
	    for (i = 0; i < N; i++) {
		I = phgSolverMapE2G(solver, k, e, i);
		for (j = 0; j <= i; j++) {
		    /* \int_T A\grad u1 . \grad v */
		    J = phgSolverMapE2G(solver, k, e, j);
#if CACHE_MODE
		    val = phgQCIntegrate(q, eno, Q_GRAD, j, q, eno, Q_AGrad, i);
#else
		    val = phgQCIntegrate(q, eno, Q_GRAD, j, q, eno, Q_GRAD, i)
			  * (k == 0 ? coeff1 : coeff2);
#endif
		    phgSolverAddGlobalMatrixEntry(solver, I, J, val); 
		    if (i != j)
			phgSolverAddGlobalMatrixEntry(solver, J, I, val); 
		}

		/* \int_T f1 v */
		val = phgQCIntegrate(q, eno, Q_BAS, i, q, eno, Q_f, 0);
		phgSolverAddGlobalRHSEntry(solver, I, val);
	    }
	}
	phgFree(rule_m);
	phgFree(rule_p);
	rule_m = rule_p = NULL;

	/*==================== face integrals ====================*/

	if (e->mark == 0) {	/* the interface */
	    FLOAT G[2] = {0.,0.};
	    phgQCSetRule(qm, rule_0, -1.);
	    phgQCSetRule(qp, rule_0, -1.);
	    do_face(xi, solver, G, qm, e, -1, /*dof_id*/0,  coeff1,
				   qp, e, -1, /*dof_id1*/1, coeff2);

	    /* Apply jump conditions on e\cap\Gamma */
	    for (i = 0; i < N; i++) {
		FLOAT a, val1 = 0.0, val2 = 0.0;

		/* \int [A*gN.n] {v} */
#if CACHE_MODE == 2
		a = (phgQCIntegrateFace(qm, eno, -1, Q_AgN_n, PROJ_NONE, 0,
					qm, eno, -1, Q_BAS,   PROJ_NONE, i) -
		     phgQCIntegrateFace(qp, eno, -1, Q_AgN_n, PROJ_NONE, 0,
					qp, eno, -1, Q_BAS,   PROJ_NONE, i));
#elif CACHE_MODE == 1
		a = (phgQCIntegrateFace(qm, eno, -1, Q_AgN, PROJ_DOT, 0,
					qm, eno, -1, Q_BAS, PROJ_NONE, i) -
		     phgQCIntegrateFace(qp, eno, -1, Q_AgN, PROJ_DOT, 0,
					qp, eno, -1, Q_BAS, PROJ_NONE, i));
#else
		a = (phgQCIntegrateFace(qm, eno, -1, Q_gN_n, PROJ_NONE, 0,
					qm, eno, -1, Q_BAS,  PROJ_NONE, i)
		     * coeff1 -
		     phgQCIntegrateFace(qp, eno, -1, Q_gN_n, PROJ_NONE, 0,
					qp, eno, -1, Q_BAS,  PROJ_NONE, i)
		     * coeff2);
#endif
		a *= 0.5;
		val1 += a;
		val2 += a;

		/* G1 \int [A*gN.n] [(A\grad v).n] */
#if CACHE_MODE == 2
		a = (phgQCIntegrateFace(qm, eno, -1, Q_AgN_n,  PROJ_NONE, 0,
					qm, eno, -1, Q_GRAD_n, PROJ_NONE, i) -
		     phgQCIntegrateFace(qp, eno, -1, Q_AgN_n,  PROJ_NONE, 0,
					qp, eno, -1, Q_GRAD_n, PROJ_NONE, i));
#elif CACHE_MODE == 1
		a = (phgQCIntegrateFace(qm, eno, -1, Q_AgN,  PROJ_DOT, 0,
					qm, eno, -1, Q_GRAD, PROJ_DOT, i) -
		     phgQCIntegrateFace(qp, eno, -1, Q_AgN,  PROJ_DOT, 0,
					qp, eno, -1, Q_GRAD, PROJ_DOT, i));
#else
		a = (phgQCIntegrateFace(qm, eno, -1, Q_gN_n,   PROJ_NONE, 0,
					qm, eno, -1, Q_GRAD_n, PROJ_NONE, i)
		     * coeff1 -
		     phgQCIntegrateFace(qp, eno, -1, Q_gN_n,   PROJ_NONE, 0,
					qp, eno, -1, Q_GRAD_n, PROJ_NONE, i)
		     * coeff2);
#endif
		a *= G[1];
		val1 += coeff1 * a;
		val2 += -coeff2 * a;

		/* -beta \int [gD] {(A\grad v).n} (func[1] := gD) */
#if CACHE_MODE == 1
		a = (phgQCIntegrateFace(qm, eno, -1, Q_gD,   PROJ_NONE, 0,
					qm, eno, -1, Q_GRAD, PROJ_DOT, i) -
		     phgQCIntegrateFace(qp, eno, -1, Q_gD,   PROJ_NONE, 0,
					qp, eno, -1, Q_GRAD, PROJ_DOT, i));
#else
		a = (phgQCIntegrateFace(qm, eno, -1, Q_gD,     PROJ_NONE, 0,
					qm, eno, -1, Q_GRAD_n, PROJ_NONE, i) -
		     phgQCIntegrateFace(qp, eno, -1, Q_gD,     PROJ_NONE, 0,
					qp, eno, -1, Q_GRAD_n, PROJ_NONE, i));
#endif
		a *= -beta * 0.5;
		val1 += coeff1 * a;
		val2 += coeff2 * a;

		/* G0 \int [gD] [v] */
		a = (phgQCIntegrateFace(qm, eno, -1, Q_gD,  PROJ_NONE, 0,
					qm, eno, -1, Q_BAS, PROJ_NONE, i) -
		     phgQCIntegrateFace(qp, eno, -1, Q_gD,  PROJ_NONE, 0,
					qp, eno, -1, Q_BAS, PROJ_NONE, i));
		a *= G[0];
		val1 += a;
		val2 += -a;

		I = phgSolverMapE2G(solver, 0, e, i);
		phgSolverAddGlobalRHSEntry(solver, I, val1);

		I = phgSolverMapE2G(solver, 1, e, i);
		phgSolverAddGlobalRHSEntry(solver, I, val2);
	    }
	}
	phgFree(rule_0);
	rule_0 = NULL;

	/* the faces of the element */
	for (face = 0; face < NFace; face++) {
	    FLOAT nv[Dim];
	    ELEMENT *e1 = phgGetNeighbour(g, e, face);
	    INT e1no = e1 == NULL ? -1 : e1->index;
	    int face1 = -1;
	    if (e1 != NULL) {
		/* a face is only processed by one of the two neighbours with
		 * smaller index to avoid double counting/redundant computing */
		if (GlobalElement(g, eno) > GlobalElement(g, e1no))
		    continue;
		face1 = phgOppositeFace(g, e, face, e1);
	    }
	    phgGeomGetFaceOutNormal(g, e, face, nv);
	    if (e->mark == 0)
		phgQuadInterfaceFace(xi->ls, xi->ls_grad, e, face, order + 3,
				     &rule_m, NULL, &rule_p);
	    else if (e->mark < 0)
		phgQuadInterfaceFace(NULL, xi->ls_grad, e, face, order,
				     &rule_m, NULL, NULL);
	    else
		phgQuadInterfaceFace(NULL, xi->ls_grad, e, face, order,
				     NULL, NULL, &rule_p);

	    if (e->mark <= 0) {
		phgQCSetRule(qm, rule_m, -1.);
		phgQCSetConstantNormal(qm, nv);
		do_face(xi, solver, NULL, qm, e,  face,  0, coeff1,
					  qm, e1, face1, 0, coeff1);
	    }

	    if (e->mark >= 0) {
		phgQCSetRule(qp, rule_p, -1.);
		phgQCSetConstantNormal(qp, nv);
		do_face(xi, solver, NULL, qp, e,  face,  1, coeff2,
					  qp, e1, face1, 1, coeff2);
	    }

	    phgFree(rule_m);
	    phgFree(rule_p);
	    rule_m = rule_p = NULL;
	}
    } ForAllElementsEnd

    minfrac[0] = minfrac[1] = 2.0;
    ecnts[0] = ecnts[1] = ecnts[2] = 0;
    for (ii = 0; ii < phgMaxThreads; ii++) {
	if (minfrac[0] > tmp[0][ii])
	    minfrac[0] = tmp[0][ii];
	if (minfrac[1] > tmp[1][ii])
	    minfrac[1] = tmp[1][ii];
	ecnts[0] += etmp[0][ii];
	ecnts[1] += etmp[1][ii];
	ecnts[2] += etmp[2][ii];
    }
#if USE_MPI
    if (g->nprocs > 1) {
	double tmp[2];
	INT etmp[3];
	tmp[0] = minfrac[0];
	tmp[1] = minfrac[1];
	MPI_Reduce(tmp, minfrac, 2, MPI_DOUBLE, MPI_MIN, 0, g->comm);
	etmp[0] = ecnts[0];
	etmp[1] = ecnts[1];
	etmp[2] = ecnts[2];
	MPI_Reduce(etmp, ecnts, 3, PHG_MPI_INT, MPI_SUM, 0, g->comm);
    }
#endif	/* USE_MPI */
    if (minfrac[0] > 1.5)
	minfrac[0] = 0.0;
    if (minfrac[1] > 1.5)
	minfrac[1] = 0.0;
    phgPrintf("    # elements in -/0/+ regions: %"dFMT "/%"dFMT"/%"dFMT"\n",
			ecnts[0], ecnts[1], ecnts[2]);
    phgPrintf("    Minimum -/+ fractions in interface elements: %lg/%lg\n",
			minfrac[0], minfrac[1]);

    phgQCFree(&qm);
    phgQCFree(&qp);

    /* set the diagonal entry of empty rows to 1 */
    process_empty_rows(solver);

    if (dump_mat) {
	phgMatDumpMATLAB(solver->mat, "A", "A.m");
	phgPrintf("    coefficient matrix saved in \"A.m\".\n");
    }
    if (dump_rhs) {
	phgVecDumpMATLAB(solver->rhs, "b", "b.m");
	phgPrintf("    RHS saved in \"b.m\".\n");
    }
}

int
main(int argc, char *argv[])
{
    char *fn = "../test/cube.dat";
    INT refine = 0;
    GRID *g;
    XFEM_INFO *xi;
    DOF *u1_h, *u2_h, *f1_h, *f2_h, *error1, *gerror1, *error2, *gerror2, *gu_h;
    SOLVER *solver;
    size_t mem_peak;
    double t0, t;
    FLOAT L2norm, H1norm, L2err, H1err;
    BOOLEAN vtk = FALSE;

    phgOptionsRegisterFilename("-mesh_file", "Mesh file", &fn);
    phgOptionsRegisterInt("-refine", "Refinement", &refine);
    phgOptionsRegisterFloat("-r", "Radius of the sphere", &r);
    phgOptionsRegisterFloat("-xc", "Center of the sphere: x-coordinate", &xc);
    phgOptionsRegisterFloat("-yc", "Center of the sphere: y-coordinate", &yc);
    phgOptionsRegisterFloat("-zc", "Center of the sphere: z-coordinate", &zc);
    phgOptionsRegisterFloat("-coeff1", "The coefficient coeff1", &coeff1);
    phgOptionsRegisterFloat("-coeff2", "The coefficient coeff2", &coeff2);
    phgOptionsRegisterFloat("-beta", "The parameter beta", &beta);
    phgOptionsRegisterFloat("-gamma0", "The parameter gamma0", &gamma0);
    phgOptionsRegisterFloat("-gamma1", "The parameter gamma1", &gamma1);
    phgOptionsRegisterNoArg("-vtk", "Create \"interface.vtk\".", &vtk);

    phgOptionsRegisterNoArg("-dump_matrix", "Write matrix to 'A.m'", &dump_mat);
    phgOptionsRegisterNoArg("-dump_rhs", "Write RHS to 'b.m'", &dump_rhs);

    phgOptionsPreset("-dof_type DG1 -solver gmres");
#if USE_MUMPS
    phgOptionsPreset("-solver mumps -mumps_symmetry unsym");
#endif
#if USE_HYPRE
    /* BoomerAMG seems to work pretty well */
    phgOptionsPreset("-solver hypre -hypre_solver gmres -hypre_pc boomeramg");
#endif

    phgInit(&argc, &argv);

    g = phgNewGrid(-1);
    if (!phgImport(g, fn, TRUE))
	phgError(1, "can't read file \"%s\".\n", fn);

    /* refine the mesh */
    phgPrintf("Refining grid: ");
    while (refine > 0) {
	phgRefineAllElements(g, refine >= 3 ? 3 : refine);
    	phgBalanceGrid(g, 1.1, 1, NULL, 1.0);
	refine -= 3;
    }
    phgPrintf("%"dFMT" elements\n", g->nelem_global);

    phgSetupHalo(g, HALO_FACE);

#if 0
ELEMENT *e;
ForAllElements(g, e) {
    phgInfo(-1, "elem %d:\n", e->index);
    for (int i = 0; i < P8EST_FACES; i++) {
	ELEMENT *e1 = phgGetNeighbour(g, e, i);
	if (e1 != NULL)
	    phgInfo(-1, "\tneigh %d: %d, face %d\n",
				i, e1->index, phgOppositeFace(g, e, i, e1));
    }
}
phgFreeGrid(&g);
phgFinalize();
exit(0);
#endif

    xi = phgXFEMInit(g, ls_func, 2, ls_grad);

    u1_h = phgDofNew(g, DOF_DEFAULT, 1, "u1_h", DofInterpolation);
    u2_h = phgDofNew(g, DOF_DEFAULT, 1, "u2_h", DofInterpolation);

    f1_h = phgDofNew(g, DOF_DEFAULT, 1, "f1_h",  func_f1);
    f2_h = phgDofNew(g, DOF_DEFAULT, 1, "f2_h",  func_f2);

    t0 = phgGetTime(NULL);
    solver = phgSolverCreate(SOLVER_DEFAULT, u1_h, u2_h, NULL);
    solver->mat->handle_bdry_eqns = FALSE;
    phgPrintf("Building linear equations ...\n");
    t = phgGetTime(NULL);
    build_linear_system(xi, solver, u1_h, u2_h, f1_h, f2_h);
    phgPrintf("    %"dFMT" unknowns, wall time: %0.4lg\n",
		solver->rhs->map->nglobal,
		phgGetTime(NULL) - t);
    phgPrintf("Solving linear equations ...\n");
    t = phgGetTime(NULL);
    phgSolverSolve(solver, TRUE, u1_h, u2_h, NULL);
    phgPrintf("    solver: %s; nits: %d; residual: %lg; wall time: %0.4lg\n",
	solver->oem_solver->name, solver->nits, (double)solver->residual,
	phgGetTime(NULL) - t);
    phgSolverDestroy(&solver);

    t0 = phgGetTime(NULL);

    /* compute L2 and H1 errors */

    /* error := u, projected to the FE space */
    error1 = phgDofNew(g, u1_h->type, 1, "error1", func_u1);
    error2 = phgDofNew(g, u2_h->type, 1, "error2", func_u2);
    L2norm = Sqrt(phgXFEMDot(xi, error1, error2, error1, error2));

    /* gerror := Grad(u), projected to the FE space */
    gerror1 = phgDofGradient(error1, NULL, NULL, NULL);
    gerror2 = phgDofGradient(error2, NULL, NULL, NULL);
    H1norm = L2norm + Sqrt(phgXFEMDot(xi, gerror1, gerror2, gerror1, gerror2));

    /* error := u_h - error */
    phgDofAXPY(-1.0, u1_h, &error1);
    phgDofAXPY(-1.0, u2_h, &error2);
    /* L2err := |error| */
    L2err = Sqrt(phgXFEMDot(xi, error1, error2, error1, error2));

    /* gerror := Grad(u_h) - gerror */
    gu_h = phgDofGradient(u1_h, NULL, NULL, NULL);
    phgDofAXPY(-1.0, gu_h, &gerror1);
    phgDofFree(&gu_h);
    gu_h = phgDofGradient(u2_h, NULL, NULL, NULL);
    phgDofAXPY(-1.0, gu_h, &gerror2);
    phgDofFree(&gu_h);
    /* H1err := L2err + |gerror| */
    H1err = L2err + Sqrt(phgXFEMDot(xi, gerror1, gerror2, gerror1, gerror2));

    phgMemoryUsage(g, &mem_peak);
    phgPrintf("L2 error = %0.5le; H1 error = %0.5le; "
	      "mem: %0.2lfMB; time: %0.3lg\n",
		(double)L2err / (L2norm == 0. ? 1.0 : (double)L2norm),
		(double)H1err / (H1norm == 0. ? 1.0 : (double)H1norm),
		(double)mem_peak / (1024.0 * 1024.0),
		phgGetTime(NULL) - t0);

    if (vtk) {
	phgExportVTK(g, "interface.vtk", u1_h, u2_h, error1, error2, NULL);
	phgPrintf("\"interface.vtk\" created.\n"); /* FIXME: output DG to VTK */
    }

    phgDofFree(&error1);
    phgDofFree(&gerror1);
    phgDofFree(&error2);
    phgDofFree(&gerror2);


    phgDofFree(&u1_h);
    phgDofFree(&u2_h);
    phgDofFree(&f1_h);
    phgDofFree(&f2_h);

    phgXFEMFree(&xi);
    phgFreeGrid(&g);
    phgFinalize();

    return 0;
}
