/* Parallel Hierarchical Grid -- an adaptive finite element library.
 *
 * Copyright (C) 2005-2010 State Key Laboratory of Scientific and
 * Engineering Computing, Chinese Academy of Sciences. */

/* This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA */

/* $Id: ipdg.c,v 1.113 2021/06/25 13:00:59 zlb Exp $ */

#include "phg.h"
#include <string.h>
#include <math.h>

/*---------------------------------------------------------------------------
 * Sample IPDG (interior penalty discontinuous Galerkin) code
 * for solving the following Poisson equation:
 *		-\Delta u = f
 *		u|_{\Gamma} = g
 * Weak form:
 * 		a(u,v) = f(v), \forall v
 * where:
 * 	a(u,v)	= \int_\Omega \grad u . \grad v
 * 		  +
 *		  \sum_{F\in\Gamma_0\cup\Gamma} \int_F (
 *			- {\grad u}.n [v] - \beta [u]{\grad v}.n
 *			+ \gamma_0 p^2 / h_F [u][v]
 *		  )
 *		  +
 *		  \sum_{F\in\Gamma_0\cup\Gamma_N} \int_F (
 *			\gamma_1 h_F / p^2 [\grad u].n [\grad v].n 
 *		  )
 *	f(v)	= \int_\Omega f v
 *		  +
 *		  \sum_{F\in\Gamma_D} \int_F (
 *			g (-\beta(\grad v).n + \gamma0 p^2 / h_F v)
 *		  )
 *		  +
 *		  \sum_{F\in\Gamma_N} \int_F (
 *			\gamma1 h_F / p^2 g_N [\grad v].n)
 *		  )
 * where F: face, h_F: face diameter,
 * 	 \Gamma_0 is the set of interior faces,
 *	 \Gamma = \Gamma_N \cup \Gamma_D = \partial\Omega,
 *---------------------------------------------------------------------------*/

/* The IP parameters */
static FLOAT beta = -1.0, gamma0 = 1.0, gamma1 = 1.0;

static BOOLEAN dump_mat = FALSE, dump_rhs = FALSE;

static void
func_u(FLOAT x, FLOAT y, FLOAT z, FLOAT *value)
{
#if 0
    *value = Cos(2.*M_PI*x) * Cos(2.*M_PI*y) * Cos(2.*M_PI*z);
#else
    *value = Exp(x * y * z);
#endif
}

static void
func_f(FLOAT x, FLOAT y, FLOAT z, FLOAT *value)
{
#if 0
    func_u(x, y, z, value);
    *value *= 12. * M_PI * M_PI;
#else
    func_u(x, y, z, value);
    *value *= -(x * x * y * y + x * x * z * z + y * y * z * z);
#endif
}

#if 0
/* FIXME: this is noticebly (>2x) faster than calling libphg, e.g.,
 * 	./ipdg -refine 12
 * (result of IPO?) */
# include "../src/quad-cache.c"
#endif

static void
do_face(SOLVER *solver, DOF *u_h, QCACHE *qc, int Q_g, ELEMENT *e, int face)
{
    GRID *g = u_h->g;
    ELEMENT *e1;
    BYTE face1;
    int n, n1, p;   /* p = polynomial order */
    INT I, J;
    int i, j;
    FLOAT val, nv[Dim], *rule;
    FLOAT G0, G1, h, a;     /* G0 := gamma0*p^2/h, G1 := gamma1*h/p^2 */
    enum {INTERIOR, DIRICHLET, NEUMANN} bdry_flag;

    n = DofNBas(u_h, e);
    p = DofTypeOrder(u_h, e);
    h = phgGeomGetFaceDiameter(g, e, face);

    rule = phgQuadGetRule2D(g, e, face, 2 * p);
    phgQCSetRule(qc, rule, -1.);
    phgGeomGetFaceOutNormal(g, e, face, nv);
    phgQCSetConstantNormal(qc, nv);

    if ((e1 = phgGetNeighbour(g, e, face)) == NULL) {
	/* boundary face */
	bdry_flag = DIRICHLET;
	n1 = 0;
	face1 = -1;
	G0 = gamma0 * p * p / h;
	G1 = gamma1 * h / (p * (FLOAT)p);
	/* RHS */
	for (i = 0; i < n; i++) {
	    I = phgSolverMapE2G(solver, 0, e, i);
	    /* -\beta\int_\Gamma_D g (\grad v).n */
	    a = phgQCIntegrateFace(qc, e->index, face, Q_g,    PROJ_NONE, 0,
				   qc, e->index, face, Q_GRAD, PROJ_DOT,  i);
	    val = -beta * a;
	    /* G0 \int_\Gamma_D g v */
	    a = phgQCIntegrateFace(qc, e->index, face, Q_g,   PROJ_NONE, 0,
				   qc, e->index, face, Q_BAS, PROJ_NONE, i);
	    val += G0 * a;
	    phgSolverAddGlobalRHSEntry(solver, I, val);
	}
    }
    else {
	/* interior face shared by "e" and "e1" */
	bdry_flag = INTERIOR;
	face1 = phgOppositeFace(g, e, face, e1);
	n1 = DofNBas(u_h, e1);
	i = DofTypeOrder(u_h, e1);
	if (p < i)
	    p = i;
	G0 = gamma0 * p * (FLOAT)p / h;
	G1 = gamma1 * h / (p * (FLOAT)p);
    }

    /* loop on {basis funcs in e} \cup {basis funcs in e1} */
    for (i = 0; i < n + n1; i++) {
	SHORT face_i, bas_i;
	ELEMENT *e_i;

	if (i < n) {
	    e_i = e;
	    face_i = face;
	    bas_i = i;
	}
	else {
	    e_i = e1;
	    face_i = face1;
	    bas_i = i - n;
	}
	I = phgSolverMapE2G(solver, 0, e_i, bas_i);
	/* loop on {basis funcs in e} \cup {basis funcs in e1} */
	for (j = 0; j < n + n1; j++) {
	    SHORT face_j, bas_j;
	    ELEMENT *e_j;

	    if (j < n) {
		e_j = e;
		face_j = face;
		bas_j = j;
	    }
	    else {
		e_j = e1;
		face_j = face1;
		bas_j = j - n;
	    }
	    J = phgSolverMapE2G(solver, 0, e_j, bas_j);

            /* skip jumps for interior face and continuous element */
            if (DofFESpace(u_h) == FE_H1 && bdry_flag == INTERIOR)
                continue;

	    /*-----------------------------------------------------------------
	     * Note: If the normal vector is reversed, the sign of [.] changes,
	     *	      while the sign of {.} is unaffected.
	     *----------------------------------------------------------------*/

	    /* -\int {\grad u}.n [v] (i<=>v, j<=>u, n<=>e_j) */
	    a = phgQCIntegrateFace(
			qc, e_j->index, face_j, Q_GRAD, PROJ_DOT,  bas_j,
			qc, e_i->index, face_i, Q_BAS,  PROJ_NONE, bas_i);
	    if (bdry_flag == INTERIOR)
		a *= (e_i == e ? 0.5 : -0.5);
	    val = -a;

	    /* -\int \beta [u]{\grad v}.n (i<=>v, j<=>u, n<=>e_i) */
	    a = phgQCIntegrateFace(
			qc, e_j->index, face_j, Q_BAS,  PROJ_NONE, bas_j,
			qc, e_i->index, face_i, Q_GRAD, PROJ_DOT,  bas_i);
	    if (bdry_flag == INTERIOR)
		a *= (e_j == e ? 0.5 : -0.5);
	    val += -beta * a;

	    /* \int G0 [u][v] (i<=>v, j<=>u, n<=>e) */
	    a = phgQCIntegrateFace(
			qc, e_j->index, face_j, Q_BAS, PROJ_NONE, bas_j,
			qc, e_i->index, face_i, Q_BAS, PROJ_NONE, bas_i);
	    if (bdry_flag == INTERIOR && e_i != e_j)
		a = -a;
	    val += G0 * a;

	    if (bdry_flag != DIRICHLET) {
		/* \int G1 [\grad u].n [\grad v].n (i<=>v, j<=>u, n<=>e_i) */
		a = phgQCIntegrateFace(
			    qc, e_j->index, face_j, Q_GRAD, PROJ_DOT, bas_j,
			    qc, e_i->index, face_i, Q_GRAD, PROJ_DOT, bas_i);
		if (bdry_flag == INTERIOR && e_i != e_j)
		    a = -a;
		val += G1 * a;
	    }

	    phgSolverAddGlobalMatrixEntry(solver, I, J, val);
	}
    }

    phgFree(rule);
}

static void
build_linear_system(SOLVER *solver, DOF *u_h, DOF *f_h)
{
    GRID *g = u_h->g;
    DOF *dof_g = phgDofNew(g, DOF_DEFAULT, 1, "g", func_u);
    ELEMENT *e;
    QCACHE *qc;
    int Q_f, Q_g;
    
    assert(u_h->dim == 1);

    qc = phgQCNew(QD_PHG, u_h);
    Q_f = phgQCAddFEFunction(qc, f_h);
    Q_g = phgQCAddFEFunction(qc, dof_g);

#if USE_OMP
#pragma omp parallel for private(e)
#endif	/* USE_OMP */
    ForAllElementsBegin(g, e) {
	int N = DofNBas(u_h, e);
	INT I, J, eno = e->index;
	int i, j;
	FLOAT val, *rule;
	/* compute \int \grad\phi_j \cdot \grad\phi_i making use of symmetry */
	rule = phgQuadGetRule3D(g, e, 2 * DofTypeOrder(u_h, e) - 2);
	phgQCSetRule(qc, rule, -1.);
	for (i = 0; i < N; i++) {
	    I = phgSolverMapE2G(solver, 0, e, i);
	    for (j = 0; j <= i; j++) {
		J = phgSolverMapE2G(solver, 0, e, j);
		/* \int_T \grad u . \grad v */
		val = phgQCIntegrate(qc, eno, Q_GRAD, j, qc, eno, Q_GRAD, i);
		phgSolverAddGlobalMatrixEntry(solver, I, J, val); 
		if (i != j)
		    phgSolverAddGlobalMatrixEntry(solver, J, I, val); 
	    }
	    /* \int_T fv */
	    val = phgQCIntegrate(qc, eno, Q_BAS, i, qc, eno, Q_f, 0);
	    phgSolverAddGlobalRHSEntry(solver, I, val);
	}
	phgFree(rule);

	/* loop on the faces of the elements */
	for (i = 0; i < NFace; i++) {
	    ELEMENT *e1 = phgGetNeighbour(g, e, i);
	    if (e1 != NULL &&
		GlobalElement(g, e->index) > GlobalElement(g, e1->index))
		continue; /* avoid double counting and redundant computing */
	    do_face(solver, u_h, qc, Q_g, e, i);
	}
    } ForAllElementsEnd

    phgDofFree(&dof_g);
    phgQCFree(&qc);
}

int
main(int argc, char *argv[])
{
    INT refine = 0, periodicity = 0 /* {1|2|4} (1=x, 2=y, 4=z) */;
    char *fn = "../test/cube.dat";
    GRID *g;
    DOF *u_h, *f_h, *error, *gerror, *gu_h;
    SOLVER *solver;
    size_t mem_peak;
    double t0, t;
    FLOAT L2norm, H1norm, L2err, H1err;
#ifndef PHG_TO_P4EST_H
    BOOLEAN hp_test = FALSE;
    phgOptionsRegisterNoArg("-hp_test", "Test h-p DOF", &hp_test);
#endif	/* PHG_TO_P4EST_H */

    phgOptionsRegisterInt("-periodicity", "Set periodicity", &periodicity);
    phgOptionsRegisterInt("-refine", "Refinement depth", &refine);
    phgOptionsRegisterFloat("-beta", "The parameter beta", &beta);
    phgOptionsRegisterFloat("-gamma0", "The parameter gamma0", &gamma0);
    phgOptionsRegisterFloat("-gamma1", "The parameter gamma1", &gamma1);
    phgOptionsRegisterFilename("-mesh_file", "Mesh file", &fn);

    phgOptionsRegisterNoArg("-dump_matrix", "Write matrix to 'A.m'", &dump_mat);
    phgOptionsRegisterNoArg("-dump_rhs", "Write RHS to 'b.m'", &dump_rhs);

    phgOptionsPreset("-dof_type DG1 -solver gmres");

#if USE_MUMPS
    phgOptionsPreset("-solver mumps -mumps_symmetry unsym");
#endif
#if USE_HYPRE
    /* BoomerAMG seems to work pretty well */
    phgOptionsPreset("-solver hypre -hypre_solver gmres -hypre_pc boomeramg");
#endif

    phgInit(&argc, &argv);

    g = phgNewGrid(-1);
    phgSetPeriodicity(g, periodicity);

    if (!phgImport(g, fn, TRUE))
	phgError(1, "can't read file \"%s\".\n", fn);

    while (refine > 0) {
	int level = refine > 3 ? 3 : refine;
	phgRefineAllElements(g, level);
	phgBalanceGrid(g, 1.2, 0, NULL, 1.0);
	refine -= level;
    }

    phgPrintf("%"dFMT" elements, %"dFMT" submeshes, load imbalance: %lg\n",
			g->nleaf_global, g->nprocs, (double)g->lif);

    phgSetupHalo(g, HALO_FACE);

    /* The discrete solution */
#ifndef PHG_TO_P4EST_H
    if (hp_test) {
	ELEMENT *e;
	HP_TYPE *hp;
	hp = phgHPNew(g, DOF_DEFAULT->fe_space == FE_H1 ? HP_HB : HP_DG);
	ForAllElements(g, e)
	    e->hp_order = DOF_DEFAULT->order + GlobalElement(g, e->index) % 3;
	phgHPSetup(hp, FALSE);
	u_h = phgHPDofNew(g, hp, 1, "u_h", DofInterpolation);
	phgHPFree(&hp);
    }
    else 
#endif	/* PHG_TO_P4EST_H */
    {
	u_h = phgDofNew(g, DOF_DEFAULT, 1, "u_h", DofInterpolation);
    }
    phgDofSetDirichletBoundaryMask(u_h, 0);

    /* RHS function */
    f_h = phgDofNew(g, DOF_DEFAULT, 1, "f_h",  func_f);

#if 0
    /* Test DOF processing in phgSetupHalo and phgRemoveHalo */
    phgRemoveHalo(g);
    phgSetupHalo(g, HALO_FACE);
#endif

    t0 = phgGetTime(NULL);
    solver = phgSolverCreate(SOLVER_DEFAULT, u_h, NULL);
    phgPrintf("Building linear equations ...\n");
    t = phgGetTime(NULL);
    build_linear_system(solver, u_h, f_h);
    phgPrintf("    %"dFMT" DOF, %"dFMT" unknowns, wall time: %0.4lg\n",
		DofDataCountGlobal(u_h), solver->rhs->map->nglobal,
		phgGetTime(NULL) - t);
    if (dump_mat) {
	phgMatDumpMATLAB(solver->mat, "A", "A.m");
	phgPrintf("    coefficient matrix saved in \"A.m\".\n");
    }
    if (dump_rhs) {
	phgVecDumpMATLAB(solver->rhs, "b", "b.m");
	phgPrintf("    RHS saved in \"b.m\".\n");
    }
    phgPrintf("Solving linear equations ...\n");
    t = phgGetTime(NULL);
    phgSolverSolve(solver, TRUE, u_h, NULL);
    phgPrintf("    solver: %s; nits: %d; residual: %lg; wall time: %0.4lg\n",
	solver->oem_solver->name, solver->nits, (double)solver->residual,
	phgGetTime(NULL) - t);
    phgSolverDestroy(&solver);

    t0 = phgGetTime(NULL);
#ifndef PHG_TO_P4EST_H
    if (u_h->hp != NULL)
	error = phgHPDofNew(g, u_h->hp, 1, "error", func_u);
    else
#endif	/* PHG_TO_P4EST_H */
	error = phgDofNew(g, u_h->type, 1, "error", func_u);
    gerror = phgDofGradient(error, NULL, NULL, NULL);
    gu_h = phgDofGradient(u_h, NULL, NULL, NULL);
    L2norm = phgDofNormL2(error);
    H1norm = L2norm + phgDofNormL2(gerror);
    phgDofAXPY(-1.0, u_h, &error);
    phgDofAXPY(-1.0, gu_h, &gerror);
    L2err = phgDofNormL2(error);
    H1err = L2err + phgDofNormL2(gerror);
    phgMemoryUsage(g, &mem_peak);
#if 0
    phgPrintf("u_h (DOF):");
    INT i;
    for (i = 0; i < g->nelem * u_h->count_elem; i++)
	phgPrintf("%s%0.16lg", i==0 ? " (" : ", ", u_h->data_elem[i]);
    phgPrintf(")\n");
#endif
    phgPrintf("L2 error: %0.5le; H1 error: %0.5le; "
	      "mem: %0.2lfMB; time: %0.3lg\n",
		(double)L2err / (L2norm == 0. ? 1.0 : (double)L2norm),
		(double)H1err / (H1norm == 0. ? 1.0 : (double)H1norm),
		(double)mem_peak / (1024.0 * 1024.0),
		phgGetTime(NULL) - t0);
#if 0
    phgExportVTK(g, "ipdg.vtk", u_h, error);
    phgPrintf("\"ipdg.vtk\" created.\n");	/* FIXME: output DG to VTK */
#endif

    phgDofFree(&u_h);
    phgDofFree(&f_h);
    phgDofFree(&error);
    phgDofFree(&gerror);
    phgDofFree(&gu_h);

    phgFreeGrid(&g);

    phgFinalize();

    return 0;
}
