/* Parallel Hierarchical Grid -- an adaptive finite element library.
 *
 * Copyright (C) 2005-2010 State Key Laboratory of Scientific and
 * Engineering Computing, Chinese Academy of Sciences. */

/* This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA */

/* $Id: phg-to-p4est.h,v 1.25 2021/06/23 01:57:31 zlb Exp $ */

#ifndef PHG_TO_P4EST_H	/* whole file */
#define PHG_TO_P4EST_H

#include <p8est_vtk.h>
#include <p8est_iterate.h>
#include <p8est_mesh.h>
#include <phg.h>

#ifdef __cplusplus
extern "C" {
#endif

extern QDESC QD_P4EST_;
#define QD_P4EST (&QD_P4EST_)

#define p8est_topidx_t	p4est_topidx_t
#define p8est_locidx_t	p4est_locidx_t

#undef QD_PHG
#define QD_PHG	QD_P4EST

#undef NVert
#define NVert	8
#undef NEdge
#define NEdge	12
#undef NFace
#define NFace	6

extern INT p8est_i;
#undef ForAllElements
#undef ForAllElementsBegin
#undef ForAllElementsEnd
#define ForAllElements(g, e) \
	for (p8est_i = 0; \
	     p8est_i < g->nelem && (e = g->elems[p8est_i]) != NULL; \
	     p8est_i++)
#define ForAllElementsBegin(g, e) \
	for (p8est_i = 0; p8est_i < g->nelem; p8est_i++) { \
	    e = g->elems[p8est_i];
#define ForAllElementsEnd }

#undef GlobalElement
#define GlobalElement(g, no) \
	((no) < 0 ? (no) : ((no) < (g)->nelem ? \
		(no) + (g)->p8est->global_first_quadrant[(g)->rank] : \
			(g)->O2Gmap[(no) - (g)->nelem]))

/* redefine ELEMENT, stored as *p8est_quadrant_t.p.userdata */
#define ELEMENT P4EST_ELEMENT
typedef struct {
    /* Note: 'cuboid' stores the coordinates of the corners, it is to be removed
     * if we can figure out how to get coordinates of ghost elements */
    FLOAT cuboid[2][Dim];
    p8est_tree_t *tree;		/* pointer to the corresponding tree */
    p8est_quadrant_t *quad;	/* pointer to the corresponding quadrant */
    INT index, tree_id;		/* local element index */
    int mark;
} ELEMENT;

/* redefine GRID */
#define GRID P4EST_GRID
typedef struct {
    p8est_connectivity_t *conn;
    p8est_t *p8est;
    p8est_mesh_t *mesh;
    p8est_ghost_t *ghost;

    ELEMENT **elems;
    ELEMENT *ghost_data;
    INT *O2Gmap;	/* off-process (ghost) element's global indices */
    INT *partition;	/* size nprocs+1, # elements of all processes */

    FLOAT lif;

    int nprocs, rank;
    MPI_Comm comm;

#define nleaf_global	nelem_global
#define nleaf		nelem
    INT i, nelem, nelem_global, nghost;	/* iterator and # elements */
    BOOLEAN balanced; /* for only calling p8est_partition() once */
} GRID;

/* redefine DOF and DOF_TYPE */
#define DOF		P4EST_DOF
#define DOF_TYPE_	P4EST_DOF_TYPE_
#define DOF_TYPE	P4EST_DOF_TYPE

typedef struct DOF_TYPE_ {
    char *name;
    struct DOF_TYPE_ *grad_type;	/* gradient type */

    /* "points" points to list of nodes in the reference coordinates, the basis
     *  is Lagrangian iff "points" != NULL */
    FLOAT *points;	/* interpolation points (in ref. coord.) */
    FLOAT *work;	/* work area, depends on the basis type */

    /* orders != NULL => tensor product basis, with:
     * 		orders[0] the highst degree in x,
     * 		orders[1] the highst degree in y,
     * 		orders[2] the highst degree in z.
     * In this case "points" points to the following data:
     * 		points[(orders[0]+1) + (orders[1]+1) + (orders[2]+1)],
     * and the actual points are given by:
     * 		{(px[i], py[j], pz[k])},
     * 		i in [0, orders[0]], j in [0, orders[1]], k in [0, orders[2]],
     * where:
     *		px = points, py = px+orders[0]+1, pz = py+orders[1]+1,
     * and "work" points to an array of the same length as "points", with all
     * entries initialized to 0.0. */
    int *orders;

    /* dim:	space dimension of the basis (e.g. 1 for scalar, 3 for vector)
     * nbas:	# basis functions in an element
     * order:	highest polynomial order of the basis functions */
    int dim, nbas, order;

    /* finite element space */
    FE_SPACE fe_space;
} DOF_TYPE;

typedef struct {
    char *name;
    GRID *g;
    DOF_TYPE *type;
    /* data and ghost_data point respectively to dynamically allocated arrays
     * 		FLOAT[nelem][n_bas][dim],
     * 		FLOAT[nghost][n_bas][dim]
     */
    FLOAT *data, *ghost_data;
    FUNC_3D userfunc;	/* function pointer for DOF_ANALYTIC */
    int dim, poly_order;
} DOF;

#undef DOF_ANALYTIC
#define DOF_ANALYTIC	p8est_dof_analytic
#define DOF_ANALYTIC_	p8est_dof_analytic_
#define DOF_DEFAULT	p8est_dof_default

#define DOF_TYPE_ORDER_MAX	16
extern DOF_TYPE **DOF_DGQn, *DOF_ANALYTIC, *DOF_DEFAULT;

#undef DofDataCountGlobal
#define DofDataCountGlobal(u)	\
	((u)->g->nelem_global * (INT)(u)->dim * (INT)(u)->type->nbas)
#define phgDofSetDirichletBoundaryMask(u, mask)

#undef DofTypeOrder
#define DofTypeOrder(dof, e) ((dof)->type == DOF_ANALYTIC ? \
					(dof)->poly_order : (dof)->type->order)
#undef DofDim
#define DofDim(u) ((u)->dim * (u)->type->dim)
#undef DofNBas
#define DofNBas(dof, e)	(dof)->type->nbas
#undef DofFESpace
#define DofFESpace(dof) (dof)->type->fe_space

/* PHG functions are encapsulated below, with exactly the same prototypes */

#define phgQuadInterface	PHG_to_p8est_quad_interface
#define phgQuadInterfaceFace	PHG_to_p8est_quad_interface_face
#define phgQuadInterfaceMarkGrid PHG_to_p8est_mark_grid
#define phgQuadGetRule_		PHG_to_p8est_quad_get_rule_
#define phgNewGrid		PHG_to_p8est_new_grid
#define phgInit			PHG_to_p8est_init
#define phgFinalize		PHG_to_p8est_finalize
#define phgFreeGrid		PHG_to_p8est_free_grid
#define phgImport		PHG_to_p8est_import
#define phgRefineAllElements	PHG_to_p8est_refine_all_elements
#define phgBalanceGrid		PHG_to_p8est_balance_grid
#define phgSetupHalo_		PHG_to_p8est_setup_halo
#define phgSetPeriodicity(g, flags) \
	if (flags != 0) phgWarning("phgSetPeriodicity ignored.\n");
#define phgMemoryUsage		PHG_to_p8est_memory_usage
#define phgGeomGetFaceOutNormal	PHG_to_p8est_get_out_normal
#define phgGetNeighbour_	PHG_to_p8est_get_neighbour
#define phgOppositeFace		PHG_to_p8est_opposite_face
#define phgDofNew_		PHG_to_p8est_dof_new
#undef phgDofSetDataByFunction_
#define phgDofSetDataByFunction_ PHG_to_p8est_dof_set_data_by_function_
#define DOF_USER_FUNC_LAMBDA	PHG_to_p8est_DOF_USER_FUNC_LAMBDA
#define phgDofSetPolyOrder	PHG_to_p8est_dof_poly_order
#define phgDofFree		PHG_to_p8est_dof_free
#undef phgDofEval
#define phgDofEval		PHG_to_p8est_dof_eval
#define phgDofAXPBY_		PHG_to_p8est_dof_axpby
#define phgDofGradient_		PHG_to_p8est_dof_gradient
#define phgDofNormL2		PHG_to_p8est_dof_norm_L2
#define phgGeomGetDiameter	PHG_to_p8est_get_diameter
#define phgGeomGetFaceDiameter	PHG_to_p8est_get_face_diameter
#define phgGeomGetVolume	PHG_to_p8est_get_volume
#undef phgMapE2G
#define phgMapE2G		PHG_to_p8est_map_e2g
#define phgExportVTK		PHG_to_p8est_export_vtk
#define phgSolverCreate		PHG_to_p8est_solver_create
#define phgSolverSolve		PHG_to_p8est_solver_solve

GRID *phgNewGrid(int flags);
void phgInit(int *argc, char ***argv);
void phgFinalize(void);
void phgFreeGrid(GRID **gptr);
BOOLEAN phgImport(GRID *g, const char *filename, BOOLEAN distr);
void phgRefineAllElements(GRID *g, int level);
int phgBalanceGrid(GRID *g, FLOAT lif_threshold, INT submesh_threshold,
					DOF *weights, FLOAT power);
void phgSetupHalo_(GRID *g, int halo_type, BOOLEAN update_dofs);
size_t phgMemoryUsage(GRID *g, size_t *peak);
void phgGeomGetFaceOutNormal(GRID *g, ELEMENT *e, int face, FLOAT *nv);
ELEMENT *phgGetNeighbour_(GRID *g, ELEMENT *e, int face,
					const char *file, int line);
BYTE phgOppositeFace(GRID *g, ELEMENT *e, BYTE v, ELEMENT *e_op);
typedef void (*DOF_USER_FUNC_LAMBDA)(DOF *u, ELEMENT *e, int bno,
                const FLOAT *xref, FLOAT *values);
DOF *phgDofNew_(GRID *g, DOF_TYPE *type, HP_TYPE *hp, SHORT dim,
		const char *name, DOF_USER_FUNC userfunc,
		const char *srcfile, int srcline);
void phgDofSetDataByFunction_(DOF *dof, DOF_USER_FUNC func,
			  DOF_USER_FUNC_LAMBDA func_lam, FLOAT *funcvalues);
void phgDofSetPolyOrder(DOF *u, SHORT order);
void phgDofFree(DOF **pdof);
FLOAT *phgDofEval(DOF *dof, ELEMENT *e, const FLOAT x[], FLOAT *values);
DOF *phgDofAXPBY_(FLOAT a, DOF *x, FLOAT b, DOF **py,
		  const char *srcfile, int srcline, BOOLEAN check);
DOF *phgDofGradient_(DOF *src, DOF **newdof, DOF_TYPE *newtype,
		     const char *name, const char *srcfile, int srcline);
FLOAT phgDofNormL2(DOF *u);
FLOAT (*phgGeomGetCuboid(GRID *g, ELEMENT *e, FLOAT (*cuboid)[Dim]))[Dim];
INT phgQuadInterfaceMarkGrid(DOF *ls, DOF *ls_grad);
FLOAT phgGeomGetDiameter(GRID *g, ELEMENT *e);
FLOAT phgGeomGetFaceDiameter(GRID *g, ELEMENT *e, int face);
FLOAT phgGeomGetVolume(GRID *g, ELEMENT *e);
INT phgMapE2G(MAP *map, int dof_id, ELEMENT *e, int index);
const char *phgExportVTK(GRID *g, const char *fn, DOF *dof, ...);
SOLVER *phgSolverCreate(OEM_SOLVER *oem_solver, DOF *u, ...);
int phgSolverSolve(SOLVER *solver, BOOLEAN destroy, DOF *u, ...);
void phgQuadInterface(DOF *ls, DOF *ls_grad, ELEMENT *e, int order,
		      FLOAT **rule_m, FLOAT **rule_0, FLOAT **rule_p);
void phgQuadInterfaceFace(DOF *ls, DOF *grad, ELEMENT *e, int face, int order,
			  FLOAT **rule_m, FLOAT **rule_0, FLOAT **rule_p);
FLOAT *phgQuadGetRule_(GRID *g, ELEMENT *e, int which, int order);

#ifdef __cplusplus
}
#endif

#endif	/* defined(PHG_TO_P4EST_H) */
